#-----------------------------------------------------------------------------
# Name:        processmanager.py
# Purpose:     global list of processes and a list control to manage them
#
# Author:      Rob McMullen
#
# Created:     2007
# RCS-ID:      $Id: $
# Copyright:   (c) 2007 Rob McMullen
# License:     wxWidgets
#-----------------------------------------------------------------------------
"""Process management support

This module provides process management support in the form of a
global process manager.  This provides the ability to display and
control all processes from one place.

A similar class is Boa Constructor's wxPopen at http://boa-constructor.cvs.sourceforge.net/boa-constructor/boa/wxPopen.py?revision=1.6&view=markup
"""

import os, sys, struct, Queue, threading, time, socket
from cStringIO import StringIO

import wx
import wx.stc
from wx.lib.pubsub import Publisher
from wx.lib.evtmgr import eventManager
from peppy.editra.eclib.outbuff import EVT_PROCESS_START, EVT_UPDATE_TEXT, EVT_PROCESS_EXIT, EVT_PROCESS_ERROR, ProcessThread

try:
    from peppy.debug import *
except:
    def dprint(txt=""):
        print txt

    class debugmixin(object):
        debuglevel = 0
        def dprint(self, txt=""):
            if self.debuglevel > 0 and txt:
                dprint(txt)
            return True

if __name__ == "__main__":
    import __builtin__
    __builtin__._ = str


_GlobalProcessManager = None

def ProcessManager():
    global _GlobalProcessManager

    if _GlobalProcessManager is None:
        _GlobalProcessManager = _ProcessManager()
        wx.GetApp().Bind(EVT_PROCESS_EXIT, ProcessManager().OnProcessEnded)
        wx.GetApp().Bind(EVT_UPDATE_TEXT, ProcessManager().OnUpdateOutput)

    return _GlobalProcessManager


class JobOutputMixin(object):
    def setupJobCallbacks(self):
        self.Bind(EVT_PROCESS_START, self.OnProcessStart)
        self.Bind(EVT_PROCESS_ERROR, self.OnProcessStartFailure)
        self.Bind(EVT_UPDATE_TEXT, self.OnProcessStdout)
        self.Bind(EVT_PROCESS_EXIT, self.OnProcessExit)
        
    def OnProcessStartFailure(self, evt):
        dprint("Couldn't run %s" % p.cmd)

    def OnProcessStart(self, evt):
        dprint("Started process %d" % p.pid)

    def OnProcessStdout(self, evt):
        dprint("stdout: '%s'" % text)
    
    def OnProcessStderr(self, evt):
        dprint("stderr: '%s'" % text)
    
    def OnProcessExit(self, evt):
        dprint("Finished with pid=%d" % p.pid)


class JobOutputSaver(object):
    def __init__(self, finishCallback):
        self.start = None
        self.finish = None
        self.exit_code = None
        self.stdout = []
        self.stderr = []
        self.callback = finishCallback
    
    def __str__(self):
        lines = []
        if self.start:
            lines.append("Started %s" % (time.asctime(time.localtime(self.start))))
        if self.stdout:
            lines.append("stdout:")
            lines.extend(self.stdout)
        if self.stderr:
            lines.append("stderr:")
            lines.extend(self.stderr)
        if self.finish:
            lines.append("Finished %s" % (time.asctime(time.localtime(self.finish))))
        lines.append("Exit code = %s" % self.exit_code)
        return os.linesep.join(lines)
    
    def getErrorText(self):
        return os.linesep.join(self.stderr)
    
    def startupFailureCallback(self, p):
        self.callback(self)

    def startupCallback(self, p):
        self.starttime = time.time()

    def stdoutCallback(self, p, text):
        self.stdout.append(text)
        
    def stderrCallback(self, p, text):
        self.stderr.append(text)
        
    def finishedCallback(self, p):
        self.finish = time.time()
        self.exit_code = p.exit_code
        self.callback(self)


class Job(debugmixin):
    debuglevel = 0
    
    # Use the unicode rightward double arrow as a delimiter
    arrow = u"\u21d2 "
    started = _("Started %s on")
    cwd = _("cwd = ")
    exit = _("exit code = %s")
    finished = _("Finished %s on")
    
    def __init__(self, cmd, working_dir, job_output):
        self.pid = None
        self.process = None
        self.cmd = cmd
        self.working_dir = working_dir
        self.jobout = job_output
        self.handler = wx.GetApp()
        self.stdout = None
        self.stderr = None
        self.exit_code = 0
        self.job_thread = None
    
    @classmethod
    def matchCwd(cls, line):
        """Evaluate if the line contains an embedded working directory specifier
        
        """
        if line.startswith(cls.arrow):
            start = len(cls.arrow)
            #dprint(line[start:])
            if line[start:].startswith(_(cls.cwd)):
                cwd = line[start + len(cls.cwd):].strip()
                return cwd
        return ""
    
    def getStartMessage(self):
        return "%s%s %s\n%s%s%s\n" % (self.arrow, self.started % self.cmd, time.asctime(time.localtime(time.time())), self.arrow, self.cwd, self.working_dir)

    def getFinishMessage(self):
        return "%s%s\n%s%s %s" % (self.arrow, self.exit % self.exit_code, self.arrow, self.finished % self.cmd, time.asctime(time.localtime(time.time())))

    def run(self, text=""):
        assert self.dprint("Running %s in %s" % (self.cmd, self.working_dir))
        self.job_thread = ProcessThread(self.jobout, self.cmd, cwd=self.working_dir)
        self.job_thread.start()

    def kill(self):
        assert self.dprint()
        if self.job_thread is not None:
            self.job_thread.abort = True

    def isRunning(self):
        return self.job_thread is not None

    def OnCleanup(self, evt):
        """Cleanup handler on process exit.

        This must be called as part of the event loop, otherwise the
        process doesn't exist and it causes crashes in wx when trying
        to read the last bit of data from the streams.
        """
        assert self.dprint()
        self.readStreams()
        self.exit_code = evt.GetExitCode()
        if self.process and self.process.Exists(self.pid):
            self.process.Destroy()
            self.process = None
        wx.CallAfter(self.jobout.finishedCallback, self)

class _ProcessManager(debugmixin):
    """Display a list of all subprocesses.
    """
    debuglevel = 0
    autoclean = True
    timer = None
    
    jobs = []
    job_lookup = {}

    def idle(self):
        for job in self.jobs:
            if job.process:
                job.readStreams()
    
    def OnUpdateOutput(self, evt):
        self.idle()
    
    def run(self, cmd, working_dir, job_output, input=""):
        job = Job(cmd, working_dir, job_output)
        self.jobs.append(job)
        assert self.dprint("Running job %s" % cmd)
        job.run(input)
        if job.pid > 0:
            self.job_lookup[job.pid] = job
            Publisher().sendMessage('peppy.processmanager.started', job)
        if not self.timer:
            self.__class__.timer = wx.Timer(wx.GetApp())
            self.__class__.timer.Start(500)
        return job

    def kill(self, pid):
        job = self.lookup(pid)
        if job:
            job.kill()

    def lookup(self, pid):
        if pid in self.job_lookup:
            return self.job_lookup[pid]
        return None

    def getJobs(self):
        # return a copy!
        return [j for j in self.jobs]

    def OnProcessEnded(self, evt):
        pid = evt.GetPid()
        assert self.dprint("process ended! pid=%d" % pid)
        job = self.lookup(pid)
        if job:
            job.OnCleanup(evt)
            wx.CallAfter(self.cleanup, job)
        # evt.Skip() # Don't call Skip here, or it causes a crash!

    def cleanup(self, job):
        assert self.dprint("in cleanup")
        if self.autoclean:
            self.removeJob(job)
        if not self.jobs:
            self.__class__.timer.Stop()
            self.__class__.timer = None
        self.finished(job)

    def removeJob(self, job):
        if job.pid in self.job_lookup:
            assert self.dprint("removing pid=%d" % job.pid)
            del self.job_lookup[job.pid]
            jobs = [j for j in self.jobs if j != job]
            self.jobs = jobs

    def finished(self, job):
        assert self.dprint("sending peppy.processmanager.finished")
        Publisher().sendMessage('peppy.processmanager.finished', job)
        

class ProcessList(wx.ListCtrl, debugmixin):
    """Display all processes
    """

    def __init__(self, parent, **kwargs):
        wx.ListCtrl.__init__(self, parent, style=wx.LC_REPORT, **kwargs)
        self.createColumns()

        #self.Bind(wx.EVT_IDLE, self.OnIdle)
        self.Bind(wx.EVT_LIST_ITEM_ACTIVATED, self.OnItemActivated)
        Publisher().subscribe(self.msgStarted, 'peppy.processmanager.started')
        Publisher().subscribe(self.msgFinished, 'peppy.processmanager.finished')
        self.reset()

    def OnIdle(self, evt):
        ProcessManager().idle()
        evt.Skip()

    def msgStarted(self, msg):
        job = msg.data
        self.reset()

    def msgFinished(self, msg):
        job = msg.data
        self.reset()

    def createColumns(self):
        self.InsertColumn(0, "PID")
        self.InsertColumn(1, "Running")
        self.InsertColumn(2, "Cmd")

    def reset(self):
        pm = ProcessManager()
        list_count = self.GetItemCount()
        jobs = pm.getJobs()
        index = 0
        for job in jobs:
            if index >= list_count:
                self.InsertStringItem(sys.maxint, str(job.pid))
            else:
                self.SetStringItem(index, 0, str(job.pid))
            self.SetStringItem(index, 1, str(job.isRunning()))
            self.SetStringItem(index, 2, job.cmd)

            index += 1

        if index < list_count:
            for i in range(index, list_count):
                # don't have to increment the index to DeleteItem
                # because the list gets shorter by one each time.
                self.DeleteItem(index)

        if index > 0:
            self.ResizeColumns()

    def ResizeColumns(self):
        for col in range(self.GetColumnCount()):
            self.SetColumnWidth(col, wx.LIST_AUTOSIZE)

    def OnItemActivated(self, evt):
        index = evt.GetIndex()
        pid = int(self.GetItem(index, 0).GetText())
        p = ProcessManager().lookup(pid)
        assert self.dprint(p)
        evt.Skip()

    def kill(self):
        pm = ProcessManager()
        index = self.GetFirstSelected()
        while index != -1:
            assert self.dprint("killing job %d" % (index, ))
            pid = int(self.GetItem(index, 0).GetText())
            pm.kill(pid)
            index = self.GetNextSelected(index)


if __name__ == '__main__':
    class jobframe(wx.Frame, JobOutputMixin):
        def __init__(self):
            wx.Frame.__init__(self, None, -1, 'Process Manager Test',
                              size=(640,480))
            usercommand = True
            
            mainsizer = wx.BoxSizer(wx.VERTICAL)
            listsizer = wx.BoxSizer(wx.HORIZONTAL)
            sizer = wx.BoxSizer(wx.VERTICAL)
            
            if usercommand:
                b = wx.Button(self, -1, "Start User Command")
                sizer.Add(b, 0, wx.EXPAND|wx.ALL, 2)
                b.Bind(wx.EVT_BUTTON, self.OnStartCmd)
                
                h = wx.BoxSizer(wx.HORIZONTAL)
                cmdlabel = wx.StaticText(self, -1, 'User Command')
                h.Add(cmdlabel, 0, wx.EXPAND, 0)
                self.cmd = wx.TextCtrl(self, -1, '')
                h.Add(self.cmd, 1, wx.EXPAND, 0)
                mainsizer.Add(h, 0, wx.EXPAND, 0)
                
                self.cmd.SetValue("python -u")
            
            b = wx.Button(self, -1, "Start Sample Python Loop")
            sizer.Add(b, 0, wx.EXPAND|wx.ALL, 2)
            b.Bind(wx.EVT_BUTTON, self.OnStart)
            
            b = wx.Button(self, -1, "Start Sample Python Loop: CR, no LF")
            sizer.Add(b, 0, wx.EXPAND|wx.ALL, 2)
            b.Bind(wx.EVT_BUTTON, self.OnStartCRnoLF)
            
            b = wx.Button(self, -1, "Kill")
            sizer.Add(b, 0, wx.EXPAND|wx.ALL, 2)
            b.Bind(wx.EVT_BUTTON, self.OnKill)
            
            #b = wx.Button(self, -1, "Cleanup")
            #sizer.Add(b, 0, wx.EXPAND|wx.ALL, 2)
            #b.Bind(wx.EVT_BUTTON, self.OnCleanup)
            
            b = wx.Button(self, -1, "Quit")
            sizer.Add(b, 0, wx.EXPAND|wx.ALL, 2)
            b.Bind(wx.EVT_BUTTON, self.OnQuit)
            
            self.list = ProcessList(self)
            listsizer.Add(sizer, 0, wx.EXPAND|wx.ALL, 0)
            listsizer.Add(self.list, 1, wx.EXPAND|wx.ALL, 0)
            
            mainsizer.Add(listsizer, 1, wx.EXPAND|wx.ALL, 0)

            self.out = wx.TextCtrl(self, -1, '',
                                   style=wx.TE_MULTILINE|wx.TE_READONLY|wx.TE_RICH2)
            
            mainsizer.Add(self.out, 1, wx.EXPAND|wx.ALL, 10)

            self.SetSizer(mainsizer)
            self.SetAutoLayout(True)
        
        def OnProcessStdout(self, evt):
            dprint()
            self.out.AppendText("job=%s text=%s" % ("unknown", evt.GetValue()))
    
        def OnStartCmd(self, evt):
            p = ProcessManager().run(self.cmd.GetValue(), os.getcwd(), self)
            dprint("OnStart: pid=%d" % p.pid)
            
        def OnStart(self, evt):
            p = ProcessManager().run("python -u", os.getcwd(), self, """\
import time

for x in range(10):
    print 'loop #%d' % x
    time.sleep(1)
""")
            dprint("OnStart: p=%s" % p)
            
        def OnStartCRnoLF(self, evt):
            p = ProcessManager().run("python -u", os.getcwd(), self, """\
import time

for x in range(10):
    print 'loop #%d\\r' % x,
    time.sleep(1)
""")
            dprint("OnStart: pid=%d" % p.pid)
            
        def OnKill(self, evt):
            print "kill highlighted process"
            self.list.kill()

        def OnCleanup(self, evt):
            print "cleanup list & remove completed jobs"
            ProcessManager().cleanup()

        def OnQuit(self, evt):
            sys.exit(1)
            

    def test():
        app = wx.PySimpleApp()
        f = jobframe()
        f.Show()

        app.MainLoop()

    test()
